
function hanleMermaidError(err)
{
	console.log("mermaid:")
	console.log(err);
	return true;
}

function trimstring(str, ch)
{
	let start = 0, end = str.length;
	while(start < end && str[start] === ch)
		++start;

	while(end > start && str[end - 1] === ch)
		--end;

	return (start > 0 || end < str.length) ? str.substring(start, end) : str;
}

const filterList = ["AND", "OR", "NOT", "and", "or", "not", "?", "*", "-", "&", "\\", "/"];
function performMark(search)
{
	const searchterms = search
	.replaceAll('"', '')
	.replaceAll('*', '')
	.split(" ")
	.filter(word => !filterList.includes(word))
	.join(" ");	
	
	var options = {
	"className": "highlight",
	"caseSensitive": false,
	"ignoreJoiners": true,
	"acrossElements": false,
	"separateWordSearch": true,
	"diacritics": true
	};

	var markInstance = new Mark(document.querySelector("#PageContent"));
	  markInstance.unmark({
		done: function(){
			markInstance.mark(trimstring(searchterms, '"'), options);
		}
	  });
}

function GetContent(name) // onclick on div
{
	window.location.hash = "#" + name;
	FindMenuItem(name);
}
function GoForward2Page()
{
	let el = document.querySelector("#historyBackLink");
	let newlocation = Number(el.getAttribute("data-currentlocation")) - 1;
	el.setAttribute("data-currentlocation", newlocation);
	Go2Page(newlocation);
	return false;
}
function GoBack2Page()
{
	let el = document.querySelector("#historyBackLink");
	let newlocation = Number(el.getAttribute("data-currentlocation")) + 1;
	el.setAttribute("data-currentlocation", newlocation);
	Go2Page(newlocation);
	return false;
}
async function Go2Page(newlocation)
{
	let visitedPages = JSON.parse(localStorage.getItem('visitedPages')) || [];
	let first = 0;
	let last = visitedPages.length - 1;
	if(last < 0)
	{
		alert("history empty");
		return;
	}
	if(newlocation >= 0 && newlocation <= last)
	{
		let vp = visitedPages[last - newlocation];
		if(vp) {
			localStorage.setItem('visitedPages', JSON.stringify(visitedPages));
			__GetContentPage(vp);
			FindMenuItem(vp.replaceAll("%20", " ")); //todo			
		}
	} else {
		// alert(`no more history: ${first}, ${last - newlocation}, ${last}`);
	}
}

function GetContentPage(name, hash="") // hashchanged event
{
	if(name==='') return;
	let visitedPages = JSON.parse(localStorage.getItem('visitedPages')) || [];
	let lastElement = visitedPages[visitedPages.length - 1];
	if(lastElement !== name)
	{
		visitedPages.push(name);
		localStorage.setItem('visitedPages', JSON.stringify(visitedPages));
	}
	__GetContentPage(name, hash);
				
}

const brRegex = /<br\s*\/?>/gi;
async function __GetContentPage(name, hash) 
{
	name = decodeURI(name);
	document.title = name.substr(name.lastIndexOf('/')+1).replace(/(\.md)+$/, "");
	let rsb = document.getElementById("RightSideBar");
	if(rsb!=null) rsb.innerHTML = "";;
				
	let cp = document.getElementById("PageContent");
	if(typeof(GetContentPageBC) === typeof(Function) )
	{
		// GetContentPageBC(name, cp, processContent);
		let html = await GetContentPageBC(name);
		cp.innerHTML = html.replace(brRegex, `<br>\r\n`);
		processContent(name, hash);
		return;
	}
	await fetch(docs + 'page/' + name) // ,{headers: {'mode':'cors', 'Accept':'text/html'}}
		.then(
			response => response.text()
		) 
		.then(html => {
		if(html)
		{
			cp.innerHTML = html.replace(brRegex, `<br>\r\n`);;
			processContent(name, hash);
		}
	  })
	  .catch((err) => 
		console.warn('Something went wrong.', err)
	  );
}

function processContent(name, hash="")
{
		run_mermaid_hljs();

		if(name.endsWith(".md"))
		{
			GetRightSideBar(name);
		}
		let sfield = document.querySelector("#Search").value;
		if(sfield!=null && sfield.value!="")
		{	
			performMark(sfield);
		}
		if(hash!=="")
		{
			scrollToHeading(hash);	
		}
		document.querySelector("html").scrollTop = 0;
}
function run_mermaid_hljs()
{
	mermaid.run(
	{
		querySelector: '.mermaid,.language-mermaid,.lang-mermaid',
		suppressErrors: false,
	}); 
	hljs.configure({ useBR: true });
	hljs.highlightAll();
	setTimeout(() => { 
			AddHoverListeners();	
		}, 500);
	MathJax.typesetPromise();
	lucide.createIcons();
}
function AddHoverListeners()
{
	let mermaids = document.querySelectorAll(".mermaid");
	mermaids.forEach(mer => {
		mer.addEventListener('click', () => {
		let enlarged = document.querySelectorAll(".enlarged");
		enlarged.forEach(enl => {
			enl.remove();
		});
		let cloned = mer.cloneNode(true);
		//cloned.style.position = "relative"
		cloned.style.top = event.currentTarget.getBoundingClientRect().y + "px";
		cloned.style.width = (1.2 * event.currentTarget.firstChild.getBoundingClientRect().width) + "px";
		cloned.style.transform = "scale(1.2)";
		cloned.classList.add("enlarged");			
		cloned.addEventListener('click', () => {
			event.currentTarget.remove();
		});
		cloned.addEventListener('mouseout', () => {
			let mytop = cloned.offsetTop;
			let mybottom = cloned.offsetTop + 0.9 * cloned.getBoundingClientRect().height;
			let myleft = cloned.getBoundingClientRect().x;
			let myright = cloned.getBoundingClientRect().x + 0.9 * cloned.getBoundingClientRect().width;
			if(event.clientY <= mytop ||
				event.clientY >= mybottom ||
				event.clientX <= myleft ||
				event.clientX >= myright
			)
			{
				event.currentTarget.remove();
			}
		});
		document.documentElement.appendChild(cloned);
		
	  });
	});
}
async function GetSourcePage(name)
{
	let rsb = document.getElementById("RightSideBar");
	rsb.innerHTML = "";
	if( typeof(GetSourcePageBC) === typeof(Function) )
	{
		let markdown = await GetSourcePageBC(name);
		CreatePage(markdown, name); 
		return false;
	}
	if(name==='') return;			
	fetch(docs + 'pagesource/' + name)
		.then(
			response => response.text()
		) 
		.then(markdown => {
			CreatePage(markdown, name); 
	  })
	  .catch((err) => 
		console.warn('Something went wrong.', err)
	  );
	  return false;
}
function CreatePage(markdown, name)
{
	let pc = document.getElementById("PageContent");
	pc.innerHTML = "";
	const ta = document.createElement("textarea");
	ta.setAttribute("id", "EditedMarkdown");
	ta.setAttribute("lang", "en");
	ta.classList.add("PageEditor");
	pc.appendChild(ta);
	ta.innerHTML = markdown;
	AttachMarkdownEditor(name, ta, "enter your markdown content..", "loading...", PostPreview);
}
async function PostChanges(name, content)
{
	if(name==='') return;
	alert('saving: ' + name);
	if( typeof(PostChangesBC) === typeof(Function) )
	{
		let result = await PostChangesBC(name, content.replace(/'/g, "''"));
		// alert(result);
		if( typeof(DataCommitBC) === typeof(Function) )
		{
			DataCommitBC();
		}	
		return;
	}

	fetch(docs + 'pagechange/' + name, 
	{
		method: "post",
		headers: {
			'Accept': 'text/html',
			'Content-Type': 'text/plain'
		},
		body: content
	}
	)
		.then(
			response => response.text()
		) 
		.then(html => {
			alert(html + " page changed");
			GetContentPage(name);
	  })
	  .catch((err) => 
		console.warn('Something went wrong.', err)
	  );								 
}

const debouncedChangedSearch = debounce(changedSearch, 500);
function changedSearch(evt)
{
	if(evt === undefined || evt.key.length === 1 || evt.key == "Enter" || evt.key === "Backspace" || evt.key === "Delete") // do not act on e.g. arrow keys
	{
		let currenttopicel = document.querySelector('#CurrentTopic');
		let isactive = currenttopicel.classList.contains('IsActive');
		_GetSearchContent(isactive);
	}
}

function ToggleSearch()
{
	let currenttopicel = document.querySelector('#CurrentTopic');
	currenttopicel.classList.toggle("IsActive");
	let isactive = currenttopicel.classList.contains('IsActive');
	let search = document.querySelector('#Search').value.trim();
	if(search !== "")
	{
		_GetSearchContent(isactive);
	}
}

function GetSearchContent(useCurrentTopic)
{
	let currenttopicel = document.querySelector('#CurrentTopic');
	let isactive = currenttopicel.classList.contains('IsActive');
	_GetSearchContent(isactive);
}

function _GetSearchContent(useCurrentTopic)
{
	const searchel = document.querySelector('#Search');
	let search = searchel.value.trim();	
	
	if(search==="")
	{
		search = getSearchParam("search"); //value in url, e.g. on browser back
		searchel.value = search;
	}
	if(search.indexOf('"') < 0)
	{
		search = search.replaceAll(" and ", " AND ").replaceAll(" or ", " OR ").replaceAll(" not ", " NOT ");
	}
	document.title = "search: " + search;
	let currenttopicel = document.querySelector('#CurrentTopic');
	let currenttopic = currenttopicel.getAttribute("data-topic").trim();
	let searchurl = window.location.pathname + "?search=" + search;
	history.replaceState({page:"search: " + search}, "search", searchurl);
	let cp = document.querySelector("#PageContent");
	let rsb = document.getElementById("RightSideBar");
	cp.innerHTML = "";
	rsb.innerHTML = "";
	if(search=="" && !currenttopicel.classList.contains('IsActive'))
	{
		history.replaceState(null, "", window.location.pathname);
		currenttopicel.innerText = "";
		currenttopicel.setAttribute("data-topic", "");
		currenttopicel.setAttribute("title", "");
		currenttopicel.classList.add("hideme");
		return;
	}
	const words = search.trim().split(/\s+/);
	if (words.length === 1) {
		search = "*" + search + "*";
	}

	let lsb = document.getElementById("LeftSideBar");
	let topicterm = useCurrentTopic ? currenttopic : "";
	if( typeof(GetSearchContentInTopicBC) === typeof(Function) )
	{
		GetSearchContentInTopicBC(search, topicterm, lsb);
		return;
	}

	fetch(docs + 'searchtoc?search=' + search + '&topic=' + topicterm) // ,{headers: {'mode':'cors', 'Accept':'text/html'}}
		.then(
			response => response.text()
		) 
		.then(r => {
			let firstfound = hideDetailsWithoutHref(lsb, r, "data-searchhit", "true");
			if(firstfound!="")
			{
				GetPartialPages(search, JSON.parse(r));
			}
	  })
	  .catch((err) => 
		console.warn('Something went wrong.', err)
	  );
}

async function GetPartialPages(search, searchresult)
{	
	let cp = document.getElementById("PageContent");
	cp.innerHTML = "";
	for (let [index, pp] of searchresult.entries())
	{
		let islast =  index === searchresult.length - 1;
		let href = pp.href.substr(1);
		await GetPartialPage(cp, href, false, pp.order, pp.score, searchresult.length, pp.nhits);
		if(islast)
		{			
			run_mermaid_hljs();
			performMark(search);
			AddSearchresultScrollbarListener();
			AddTitleClickListeners(cp);
		}
	}
}

async function LoadSearchedPage(searchResult = event.target.closest('div.searchresult'))
{
	if(!searchResult.classList.contains("fullpage"))
	{
		let anchor = searchResult.querySelector("div.searchresult > a:has(.load-page-inline)");
		let href = anchor.getAttribute("href").substr(1);
		searchResult.classList.add("fullpage");
		let resultcontent = searchResult.querySelector("div.resultcontent");
		await GetPartialPage(resultcontent, href, true, 0,0,0,0);
		run_mermaid_hljs();
		let sfield = document.querySelector("#Search").value;
		if(sfield!=null && sfield.value!="")
		{
			performMark(sfield);
		}	
	}
	return false;
}

function HighlightMenuItem(href)
{
	let lsb = document.getElementById("LeftSideBar");
	removeNestedAttributes(lsb, "a", "data-active");
	let item = lsb.querySelector(`a[href='#${href}']`);
	item.setAttribute("data-active", "true");
	item.scrollIntoView({
			behavior: 'smooth', 
			block: 'start' // or nearest , inline:start
		});
}

function ShowSearchResult(evt, showme=true)
{
	let el = evt.target;
	let searchresult = el.closest('.searchresult');
	let sr = searchresult.querySelector('.searchresult-border-title');
	if(showme)
	{
		sr.classList.add("ShowSearchResult");

	} else {
		setTimeout(function() {
			sr.classList.remove("ShowSearchResult");
			}, 500
		);
	}
}

function toggleExpanded(evt, href)
{
	let searchresult = evt.target.closest('.searchresult');
	/* this would reset expanded searchresults, confusing since the position of the newly clicked element changes
	let pc = searchresult.closest('#PageContent');
	let allsr = pc.querySelectorAll('.searchresult');
	allsr.forEach((sr) => {
		sr.classList.remove('expanded');
	});
	*/

	if(searchresult !== undefined)
	{
		let org_scrollHeight = searchresult.scrollHeight;
		searchresult.classList.toggle('expanded');
		let isexpanded = searchresult.classList.contains('expanded');
		if(!isexpanded)
		{
			searchresult.scrollTop = 0;
			//GetRightSideBar("");
		}
		else
		{
			//GetRightSideBar(href); // todo: shows sidebar but scolltoelement confusing
		}
		const smallpage = searchresult.scrollHeight <= searchresult.clientHeight;
		searchresult.style.height = '';
		if(isexpanded && smallpage) {
			searchresult.style.height = (0.9*org_scrollHeight) + "px";
		}		
	}
}

async function GetPartialPage(cp, href, fullpage=false, order, score, nshown, nhits)
{
	return fetch(docs + (fullpage?'page/':'partialpage/') + href)
		.then(
			response => response.text()
		) 
		.then(html => {
			if(fullpage)
			{
				cp.innerHTML = html.replace(brRegex, '<br>\r\n');
				AddTitleClickListener(cp, href);
			}
			else
			{
				let fullhref = _removeSearchParam(window.location.href) + "#" + href;
				let resulttxt = nhits > nshown ? 
				`${order}-${nshown} of ${nhits} results` :
				`${order}-${nshown} results`
				;
				let breadcrumb = `${href.substr(0, href.lastIndexOf('/')).replaceAll('/', '  >  ')}`;
				let scoretxt =  `score: ${score}`;
				
				cp.innerHTML += `
				<div class="searchresult" 
					onmouseover="ShowSearchResult(event);HighlightMenuItem('${href}')"
					onmouseenter="ShowSearchResult(event);"
					onmouseleave = "ShowSearchResult(event, false)"
					ondblclick="toggleExpanded(event, '${href}')">
					<span class="breadcrumb">${breadcrumb}</span>
					<a href="#${href}" onclick="toggleExpanded(event, '${href}'); return false;" title="Explore">
						<span class="load-page-inline" data-lucide="square-arrow-out-down-left"></span>
					</a>
					<a href="${fullhref}" title="Open page">
						<span class="open-page" data-lucide="square-arrow-out-up-right"></span>
					</a>
				<div class="resultcontent">` + html.replace(brRegex, `<br>\r\n`) + `</div>
				<div class="searchresult-border-title" title="${scoretxt}">${resulttxt}</div>
				</div>`;				
			}
			
	  })
	  .catch((err) => 
		console.warn('Something went wrong.', err)
	  );	
}

function AddTitleClickListeners(el) // searchResult.querySelector('a:has(svg.open-page)')
{
	const searchResult = el.querySelectorAll(".searchresult");
	searchResult.forEach((searchResult) => {
		let href = searchResult.querySelector("a:has(.open-page)").getAttribute("href");
		AddTitleClickListener(searchResult,href);
	});
}
function AddTitleClickListener(el, href)
{
	const h1 = el.querySelector('h1');
	h1.addEventListener('click', async () => {
		window.location = _removeSearchParam(href);
	});
}

function AddSearchresultScrollbarListener()
{
	const searchResults = document.querySelectorAll('.searchresult');

	searchResults.forEach((searchResult) => {
		searchResult.addEventListener('scroll', async () => {
		  const scrollTop = searchResult.scrollTop;
		  const clientHeight = searchResult.clientHeight;
		  const p = searchResult.scrollHeight - scrollTop;
		  const isAtBottom = p <= clientHeight + 20;
					
		  if (isAtBottom) {
			LoadSearchedPage(searchResult);
		  }
		});
	});

}

function AddSearchfieldChangeListener()
{
	let queryField = document.querySelector('#Search');
	let lsb = document.querySelector('#LeftSideBar');
	if (!queryField.dataset.listenerAdded) {
        queryField.addEventListener('change', KeepRecentSearches);
        queryField.dataset.listenerAdded = 'true';
    }
	let currenttopicel = document.querySelector("#CurrentTopic");
	let allpinicons = document.querySelectorAll('.pinicon');
	allpinicons.forEach(pinicon => {
		if (!pinicon.dataset.listenerAdded) {
			pinicon.addEventListener('click', function(event) {
				event.preventDefault(); // seems still to open catalog
				event.stopPropagation(); // seems still to open catalog
				let allpreviouslytoggled = lsb.querySelectorAll("span.pinicon.toggled");
				pinicon.classList.toggle('toggled'); 
				let myanchor = pinicon.previousElementSibling;
				let	mydetail = myanchor.closest("details");
				let mydetails = mydetail.querySelectorAll("details");
				if(pinicon.classList.contains('toggled'))
				{					
					allpreviouslytoggled.forEach(t => {
						t.classList.remove("toggled");
					});			
					currenttopicel.classList.add("IsActive");
					currenttopicel.classList.remove("hideme");
					ShowCurrentTopic(myanchor, true);
					let alldetails = lsb.querySelectorAll("details");
					alldetails.forEach(detail => {
						detail.classList.add('hideme');
					});
					mydetails.forEach(detail => {
						detail.classList.remove('hideme');
					});
					openAllParentDetails(myanchor.closest("details"));
				}
				else
				{
					currenttopicel.classList.remove("IsActive");
					currenttopicel.classList.add("hideme");
					let alldetails = lsb.querySelectorAll("details");
					alldetails.forEach(detail => {
						detail.classList.remove('hideme');
					});
					GetSearchContent(false);
				}
			});
		}		
	});
}

window.recentSearches = [];
function KeepRecentSearches()
{
	return; //todo rewrite to add searches to back-forward history, using ?search=.. etc
	let queryField = document.querySelector('#Search');
	const search = queryField.value;
	if (!window.recentSearches.includes(search)) {
		window.recentSearches.push(search);
		if (window.recentSearches.length > 10) {
				window.recentSearches.shift(); // Remove the oldest value
		}
		let dataList = document.querySelector('#recentSearches');
		dataList.innerHTML = '';
		const firstopt = document.createElement('option');
		firstopt.value = '';
		firstopt.innerHTML = '-clear-';
		dataList.appendChild(firstopt);
		window.recentSearches.slice().reverse().forEach(value => {
			if(value !== '')
			{
				const opt = document.createElement('option');
				opt.value = value;
				opt.innerHTML = value;
				dataList.appendChild(opt);
			}
		});
	}
}

function GetRightSideBar(name)
{
	let rsb = document.getElementById("RightSideBar");
	rsb.innerHTML = "";
	if(name==="") return;
	if( typeof(GetRightSideBarBC) === typeof(Function) )
	{
		GetRightSideBarBC(name, rsb);
		lucide.createIcons();
		return;
	}
	fetch(docs + 'sidetoc/' + name) // ,{headers: {'mode':'cors', 'Accept':'text/html'}}
		.then(response => response.text()) 
		.then(html => {
		rsb.innerHTML = html;
		lucide.createIcons();
	  })
	  .catch((err) => console.warn('Something went wrong.', err));
}

async function GetLeftSideBar(name)
{
	let lsb = document.getElementById("LeftSideBar");
	if( typeof(GetLeftSideBarBC) === typeof(Function) )
	{
		await GetLeftSideBarBC(name, lsb);
		AddSearchfieldChangeListener();
		lucide.createIcons();
		return;
	}
	return fetch(docs + 'globaltoc/') // ,{headers: {'mode':'cors', 'Accept':'text/html'}}
		.then(response => response.text()) 
		.then(html => {
		lsb.innerHTML = html;
		lucide.createIcons();
		AddSearchfieldChangeListener();
	  })
	  .catch((err) => console.warn('Something went wrong.', err));
}
function clickPress(event) {
	if (event.key == "Enter" || event.key === "Backspace" || event.key === "Delete") {
		let queryField = document.querySelector('#Search');
		const search = queryField.value;
		let currenttopicel = document.querySelector("#CurrentTopic");
		 // const selectionStart = activeElement.selectionStart;
      //const selectionEnd = activeElement.selectionEnd; //  || search.length==1 && event.key === "Backspace")
		if(!currenttopicel.classList.contains('IsActive'))
		{
			if(search=="") 
			{
				removeSearchParam();
				GetLeftSideBar('');
				return;
			}
			if(event.key === "Backspace" || event.key === "Delete")
			{
				if(search.length==1)
				{
					removeSearchParam();
					GetLeftSideBar('');
					return;
				}
				const activeElement = document.activeElement;
				const selectionStart = activeElement.selectionStart;
				const selectionEnd = activeElement.selectionEnd;
				if (selectionStart !== selectionEnd) 
				{
					removeSearchParam();
					GetLeftSideBar('');
					return;				
				}
			}
		} else
		{
			_GetSearchContent(false);
		}
	}
}


function ToggleDetails()
{
	const myanchor = event.target;
	const mydetail = myanchor.closest('details');
	mydetail.open = !mydetail.open;	
	FindDetailAndShowActive(myanchor);
	return true; // indicates to follow href
}

async function FindMenuItem(name, sametoc=false)
{
	let lsb = document.getElementById("LeftSideBar");
	let istoc = lsb.getAttribute("data-searchtoc") == null;
	if(!istoc && !sametoc)
	{
		let searchfield = document.querySelector("#Search");
		searchfield.value = "";
		await GetLeftSideBar('');
		lsb.removeAttribute("data-searchtoc");
	}
	let alldetails = lsb.querySelectorAll("details");
	alldetails.forEach(child => {
		child.open = false;
	});
	let item = lsb.querySelector(`a[href='#${name}']`);
	if(item)
	{		
		let mydetail = FindDetailAndShowActive(item);
		openAllParentDetails(item.closest("details"));
		
		mydetail.scrollIntoView({
			behavior: 'smooth', 
			block: 'start' // or nearest , inline:start
		});
		/*
		setTimeout(function() {
			document.querySelector("#PageContent").scrollTop = 0;
		}, 100
		);
		*/
	}
}

function ShowCurrentTopic(topicanchor, alsoemptysearch=false)
{
	let currenttopicel = document.querySelector("#CurrentTopic");
	let currenttopictxt = topicanchor.innerText;
	let currenttopichref = topicanchor.getAttribute("href").substr(1);
	
	if(currenttopichref.endsWith(".md")) // set topic of page
	{
		let idx = currenttopichref.lastIndexOf('/');
		if(idx >= 0)
		{
			currenttopichref = currenttopichref.substring(0, idx);
			let idx2 = currenttopichref.lastIndexOf('/', idx - 1);
			if(idx2 >= 0)
			{
				currenttopictxt=currenttopichref.substring(idx2 + 1, idx);
			}
		}
	} 
	currenttopicel.innerText = currenttopictxt;
	currenttopicel.setAttribute("data-topic", currenttopichref);
	currenttopicel.setAttribute("title", currenttopichref);
	let search = document.querySelector("#Search");
	if(search.value != "" || alsoemptysearch)
	{
		currenttopicel.classList.remove("hideme");
	} else
	{
		currenttopicel.classList.add("hideme");
	}
}

function ShowActive(myanchor = event.target)
{
	ScrollBarsToTop();
	FindDetailAndShowActive(myanchor);
	const href = myanchor.href;
	const url = new URL(href);
	const hash = url.hash;
	GetContentPage(hash.substr(1));
	window.location.hash = hash;
	return false; // indicate to not follow href
}

function FindDetailAndShowActive(myanchor = event.target)
{
	const mydetail = myanchor.closest('details');
	let topel = findTopMost(mydetail, "details");
	topel = topel.closest('toc');
	removeNestedAttributes(topel, "details", "data-active");
	removeNestedAttributes(topel, "a", "data-active");
	if(myanchor.parentElement.tagName.toLowerCase() === "summary")
	{
		mydetail.setAttribute("data-active", "true");
	} else
	{
		myanchor.setAttribute("data-active", "true");
	}
	return mydetail;
}

function removeNestedAttributes(topel, elname, attrname)
{
	const allel = topel.querySelectorAll(elname);
	allel.forEach(child => {
		child.removeAttribute(attrname);
	});
}

function findTopMost(curel, elname) 
{
    let cur = curel.closest(elname);
    while (cur.parentElement && cur.parentElement.tagName.toLowerCase() === elname) {
        cur = cur.parentElement;
    }
    return cur;
}

function openAllParentDetails(detailsElement) {
    let currentElement = detailsElement;
    while (currentElement && currentElement.tagName.toLowerCase() === 'details') {
        currentElement.setAttribute('open', 'true');
		currentElement.classList.remove('hideme');
        currentElement = currentElement.parentElement.closest('details');
    }
}

function hideAllDetails(topel) 
{
    const alldetails = topel.querySelectorAll('details');
	alldetails.forEach(detail => {
		detail.open = false;
		detail.classList.add('hideme');
	});
}


function hideDetailsWithoutHref(topel, searchdata)
{
	topel.setAttribute("data-searchtoc","true");
	hideAllDetails(topel);
	searchdata = JSON.parse(searchdata);
	if(searchdata.lenght == 0) return;
	const scoreMinMax = findMinMax(searchdata, 'score');
	const lengthMinMax = findMinMax(searchdata, 'length');
	let firstfound = "";
	const alldetails = topel.querySelectorAll('details');
	alldetails.forEach(detail => {
		const anchors = detail.querySelectorAll('ul>li>a');
		let detailhashit = false;
		anchors.forEach(anchor => {
			const href = anchor.getAttribute('href');
			const foundItem = searchdata.find(item => item.href === href);
			if (foundItem) {
				detailhashit= true;
				const score = foundItem.score;
				const order = foundItem.order;
				const length = foundItem.length;
				const rellength = length / (lengthMinMax.max - lengthMinMax.min);
				const relscore = score / (scoreMinMax.max - scoreMinMax.min);
				updateColor(anchor, 100 * rellength * relscore);
				anchor.setAttribute("data-searchhit", "true");
				if(firstfound == "") firstfound = href;
			} else {
				anchor.setAttribute("data-searchhit", "false");
			}
		});
		if(detailhashit)
		{
			openAllParentDetails(detail);
		}
	});
	return firstfound;
}

function findMinMax(data, key) {
    const values = data.map(item => item[key]);
    return {
        min: Math.min(...values),
        max: Math.max(...values)
    };
}

function getColorFromScore(score) {
	// Ensure the score is between 0 and 100
	score = Math.max(0, Math.min(100, score));

	// Calculate the red and green components
	const red = Math.floor((255 * score) / 100);
	const green = Math.floor((50 * (100 - score)) / 100);

	// Return the color as an RGB string
	return `rgba(${red}, ${green}, ${green}, 1)`;
}

function updateColor(el, score) {
	const color = getColorFromScore(score);
	const li = el.closest("li");
	// needs a css Toc li.page:before { color: var(--before-color, rgba(90, 90, 90, 1));} ==>
	li.style.setProperty('--before-color', color); 
}


function ScrollBarsToTop()
{
	document.querySelector("#PageContent").scrollTop = 0;
	document.querySelector("html").scrollTop = 0;
}
function GetScrollY()
{
	return document.querySelector("#PageContent").scrollTop +
	document.querySelector("html").scrollTop;
}
function ScrollToElement(id)
{
	ScrollBarsToTop();			
	let el = document.getElementById(id);
	if(el)
	{
		scrollParentToChild(el);
		/*
		document.getElementById(id).scrollIntoView({
			block: "start",
			behavior: 'smooth',
		});
		*/
	}
}
function scrollParentToChild(child)
{
	let parent = document.querySelector("#PageContent");
	let parentRect = parent.getBoundingClientRect();
	let childRect = child.getBoundingClientRect();
	const scrollTop = childRect.top - parentRect.top;
	parent.scrollTop += scrollTop;
	parent.scrollTop -= 10;
	child.classList.add("horizontalshake");
	setTimeout(function() {
		child.classList.remove("horizontalshake");
		}, 1700
	);
}

function AddUSoftKeywords()
{
	var kwa = hljs.getLanguage('sql').keywords.keyword;
	kwa.push('RELATE');
	kwa.push('MAKE');
	kwa.push('SURE');
	kwa.push('THAT');
	kwa.push('DO');
	kwa.push('ALLOW');
	kwa.push('CONNECTED');
	kwa.push('REMOVE');
	kwa.push('WHICH');
	kwa.push('GROUPED');
}

async function PostPreview(mde, previewpane, name, markdown)
{
	let html;
	if(typeof(GetRenderedSourcePageBC) === typeof(Function) )
	{
		let escmarkdown = markdown.replaceAll("\u201C", '"').replaceAll("\u201D", '"'); // replace “,” sql error otherwise
		escmarkdown = escmarkdown.replaceAll("\u2018", "'").replaceAll("\u2019", "'"); // replace ‘, ’
		escmarkdown = escmarkdown.replaceAll("'", "''"); // for sql with inline text
		escmarkdown = escmarkdown.replaceAll("`", "\u0060"); // otherwise sql parse errors if unbalanced
		html = await GetRenderedSourcePageBC(escmarkdown);	
		if(html)
		{
			html = html.replace(brRegex, `<br>\r\n`);
			SetCustomHtml(mde, previewpane, html);
		}		
	}
	else
	{
		html = await GetBackendHtml(docs + 'mdpreview/' + name, markdown, mde, previewpane);
	}
}

async function GetBackendHtml(url, markdown, mde, previewpane)
{
	fetch(url, 
	{
		method: "post",
		headers: {
			'Accept': 'text/html',
			'Content-Type': 'text/plain'
		},
		body: markdown
	}
	)
	.then(
		response => response.text()
	) 
	.then(html => {	
		if(html)
		{
			html = html.replace(brRegex, `<br>\r\n`);
			SetCustomHtml(mde, previewpane, html);
		}
		return html;
	})	
	.catch((err) => 
		console.warn('Something went wrong.', err)
	);		
}

function SetCustomHtml(mde, previewpane, html)
{
	previewpane.innerHTML = html;
	if(mde.isPreviewActive() || mde.isSideBySideActive())
	{
		run_mermaid_hljs();
	}	
}

const USoftTableButton = {
            name: "Table",
            action: function customFunction(editor) {
                const cm = editor.codemirror;
                let output = "<!--\n This table uses a USoft-specific markdown extension that allows fenced code blocks in table cells.\n";
				output+=" As the examples show, any whitespace inside the fenced code block delimiters (```) is preserved and displayed as part of the code block.\n";
				output+=" If the cell has more content than the code block, newlines are automatically added before and after the code block.";
				output+="-->\n";
				output+="| example | result |\n| -------- | -------- |\n";
				output+="|Example LR mermaid | text preceding diagram \`\`\`mermaid\nflowchart LR\nA-->B\nB-->C\nC-->A\n\`\`\` text following diagram|\n";
                output+="|Example fenced code | \`\`\`js\nLine1\n\tLine2\n\`\`\` |\n";
				output+="|Example list | \`\`\`\n* Line1\n1. Line2\n2. Line3\n\`\`\` |\n";
				output+='|Example xml | \`\`\`xml\n<root>\n\t<emp dep="dev">Jane</emp>\n</root>\n\`\`\` |\n';
				cm.replaceSelection(output);
            },
            className: "fa fa-table USoftTableButton",
            title: "Insert Table with fenced code",
        };
		
function debounce(func, wait) {
	let timeout;
	return function(...args) {
		const context = this;
		clearTimeout(timeout);
		timeout = setTimeout(() => func.apply(context, args), wait);
	};
}

// https://github.com/Ionaru/easy-markdown-editor?tab=readme-ov-file#editor-functions
function AttachMarkdownEditor(name, ta, placeholder, loadmsg, renderer_func)
{
	const debouncedPreviewRender = debounce(renderer_func, 300);
	var loadmsgtxt = loadmsg;
	const easyMDE = new EasyMDE({
		element: ta ,
		placeholder:  placeholder,
		previewRender: (plainText, preview) => { // Async method
			// renderer_func(easyMDE, preview, name, plainText)
			debouncedPreviewRender(easyMDE, preview, name, plainText);
		},
		sideBySideFullscreen: false,
		spellChecker: true,
		syncSideBySidePreviewScroll: true,
		toolbar: [
                'bold', 'italic', 'heading', '|',
                'quote', 'unordered-list', 'ordered-list', '|',
                'link', 'image', 'table', USoftTableButton, 'horizontal-rule',
				'|',
                {
                    name: 'save',
                    action: async function customSaveFunction(editor) {
                        var content = editor.value();
						await PostChanges(name, content);
                    },
                    className: 'fa fa-save',
                    title: 'Save your changes'
                },
				{
                    name: 'exit',
                    action: function ExitEditor(editor) {
						GetContentPage(name);
                    },
                    className: 'fa fa-times',
                    title: 'Exit without saving changes!'
                }, 
				'|', 'preview', 'side-by-side',
				{
                    name: 'fullscreen',
                    action: function(editor) {
                        editor.toggleFullScreen();
                    },
                    title: 'Toggle Fullscreen (F11), Fullscreen syncs the sides!', 
                    className: 'fa fa-arrows-alt' 
                }
				
            ]
	});
	easyMDE.toggleSideBySide();
	//if(fullscreen) easyMDE.toggleFullScreen();
}

function scrollToHeading(text) {
  if(text=="") return;
  
  text = text.replaceAll("+"," ").toLowerCase();
  let id = text.replaceAll(" ", "-").toLowerCase();
  const headings = document.querySelectorAll('h1, h2, h3, h4, h5, h6');
  
  for (const heading of headings) {
    if (heading.textContent.trim().toLowerCase() === text || heading.id === id) {
      // Scroll to the matching heading
      heading.scrollIntoView({ behavior: 'smooth', block: 'start' });
      return;
    }
  }
  console.log(`Heading ${text} id=${id} not found.`);
}

function getSearchParam(name, url=window.location.search)
{
	let searchParams = new URLSearchParams(url);
	if(searchParams.has(name))
	{
		return searchParams.get(name);	
	}
	return "";
}

function removeSearchParam()
{
	const url = new URL(window.location.href);
	url.searchParams.delete('search');
	window.history.replaceState({}, '', url.toString());
}

function _removeSearchParam(href)
{
	const url = new URL(href);
	url.searchParams.delete('search');
	return url.toString();	
}

function AddPageListeners()
{
	const events = ["pagehide", "pageshow", "unload", "load", "hashchange", "popstate", "beforeunload", "browser:document-loaded"];

	const eventHandler = async (event) => {
	  switch (event.type) {
		case "pagehide": {
			let isPersisted = event.persisted ? "persisted" : "not persisted";
			console.log(`Event: ${event.type} - ${isPersisted}`);
			break;
		}
		case "pageshow": {
			let isPersisted = event.persisted ? "persisted" : "not persisted";
			console.log(`Event: ${event.type} - ${isPersisted}`);
			if(!event.persisted)
			{
				GetContentPage(window.location.hash.substring(1));
			}
			break;
		}
		case "hashchange":
		{
			console.log(`Event: ${event.type} - ${window.location.hash} --- ${event.newURL}`);
			let pagename = window.location.hash.substring(1);
			
			let hashpos = pagename.indexOf("#");
			let hash = "";
			if(hashpos >= 0)
			{
				hash = pagename.substr(hashpos + 1);
				pagename = pagename.substr(0, hashpos);	
				window.history.back();							
				GetContentPage(pagename, hash);
			} /*else // handled by popstate?
			{
				const search = document.querySelector("#Search");
				search.value = "";
				GetContentPage(pagename);
				ScrollBarsToTop();
			}*/
		}
		break;
		case "popstate": // history changed
		{
			let pagename = window.location.hash.substring(1);
			let search = getSearchParam("search");
			if(pagename === "") 
			{
				if(search!=="")
				{
					changedSearch();
				}
			}
			else
			{
				removeSearchParam();
				if(search!=="")
				{
					await GetLeftSideBar('');
				}
				let searchfield = document.querySelector("#Search");
				searchfield.value = "";
				GetContentPage(window.location.hash.substring(1));
				FindMenuItem(pagename.replaceAll("%20", " "), true);
			}
		}
		break;
		case "beforeunload": // refresh
		{
			let pagename = window.location.hash.substring(1);
			console.log("refresh: " + pagename);
		}
		break;
		case "browser:document-loaded": { // browser control specific
			GetLeftSideBar('');
			let browserControlHistory = document.querySelector("#BrowserControlHistory");
			browserControlHistory.classList.remove("hideme"); // show history buttons
		}
		break;
		case "load": {
			await GetLeftSideBar('');
			let search = getSearchParam("search");
			if(search !== "")
			{
				let searchel = document.querySelector("#Search");
				let currenttopicel = document.querySelector('#CurrentTopic');
				searchel.value = search;
				let topic = getSearchParam("topic");
				if(topic !== "")
				{					
					let idx = topic.lastIndexOf("/");
					if(idx >=0)
					{
						currenttopicel.innerText = topic.substr(idx + 1);
					}
					else
					{
						currenttopicel.innerText = topic;
					}					
					currenttopicel.setAttribute("data-topic", topic);
					currenttopicel.setAttribute("title", topic);
					currenttopicel.classList.add('IsActive');
				}
				changedSearch(); 
				currenttopicel.classList.remove('IsActive');
			} else {
				removeSearchParam();
			}
			let pagename = window.location.hash.substring(1);
			if(pagename && pagename!== "")
			{
				FindMenuItem(pagename.replaceAll("%20", " "), true);
			}
		}
		break;
		default:
		  console.log(`Event: ${event.type}`);
		  break;
	  }
	};

	events.forEach((eventName) => window.addEventListener(eventName, eventHandler));	
}

const locationre = new RegExp("(?<base>.*?)(?:[^/]*).html", "g");
var matches = locationre.exec(window.location.pathname);
if(matches !== null)
{
	console.log(matches[1]);
	var docs = matches[1];
} else
{
	var docs = "/docs/";
}
